@@ -579,7 +579,7 @@ public interface UndertowMessages {
     IOException streamLimitExceeded();
 
     @Message(id = 185, value = "Invalid IP address %s")
-    IOException invalidIpAddress(String addressString);
+    IllegalArgumentException invalidIpAddress(String addressString);
 
     @Message(id = 186, value = "Invalid TLS extension")
     SSLException invalidTlsExt();
@@ -635,12 +635,16 @@ public interface UndertowMessages {
     @Message(id = 203, value = "Invalid ACL entry")
     IllegalArgumentException invalidACLAddress(@Cause Exception e);
 
-    @Message(id = 204, value = "Out of flow control window: no WINDOW_UPDATE received from peer within %s miliseconds")
-    IOException noWindowUpdate(long timeoutMiliseconds);
-
     @Message(id = 205, value = "Path is not a directory '%s'")
     IOException pathNotADirectory(Path path);
 
     @Message(id = 206, value = "Path '%s' is not a directory")
     IOException pathElementIsRegularFile(Path path);
+
+    @Message(id = 207, value = "Invalid SNI hostname '%s'")
+    IllegalArgumentException invalidSniHostname(String hostNameValue, @Cause Throwable t);
+
+    @Message(id = 208, value = "Failed to allocate resource")
+    IOException failedToAllocateResource();
+
 }
