@@ -76,6 +76,7 @@ import java.util.Deque;
 import java.util.Map;
 import java.util.Set;
 import java.util.TreeMap;
+import java.util.concurrent.atomic.AtomicLong;
 import java.util.concurrent.Executor;
 import java.util.concurrent.TimeUnit;
 
@@ -122,7 +123,6 @@ public final class HttpServerExchange extends AbstractAttachable {
      */
     public static final AttachmentKey<String> REMOTE_USER = AttachmentKey.create(String.class);
 
-
     /**
      * Attachment key that can be used as a flag of secure attribute
      */
@@ -336,6 +336,9 @@ public final class HttpServerExchange extends AbstractAttachable {
      */
     private InetSocketAddress destinationAddress;
 
+    private static final AtomicLong REQUEST_ID_GENERATOR = new AtomicLong(0);
+    private final String requestId = Long.toString(REQUEST_ID_GENERATOR.incrementAndGet());
+
     public HttpServerExchange(final ServerConnection connection, long maxEntitySize) {
         this(connection, new HeaderMap(), new HeaderMap(), maxEntitySize);
     }
@@ -447,6 +450,10 @@ public final class HttpServerExchange extends AbstractAttachable {
         return this;
     }
 
+    public String getRequestId() {
+        return requestId;
+    }
+
     /**
      * The original request URI. This will include the host name, protocol etc
      * if it was specified by the client.
@@ -1146,7 +1153,7 @@ public final class HttpServerExchange extends AbstractAttachable {
      * @return A mutable map of request cookies
      * @deprecated use either {@link #requestCookies()} or {@link #getRequestCookie(String)} or {@link #setRequestCookie(Cookie)} methods instead
      */
-    @Deprecated
+    @Deprecated(since="2.2.0", forRemoval=true)
     public Map<String, Cookie> getRequestCookies() {
         if (deprecatedRequestCookies == null) {
             deprecatedRequestCookies = new MapDelegatingToSet((Set<Cookie>)((DelegatingIterable<Cookie>)requestCookies()).getDelegate());
@@ -1231,7 +1238,7 @@ public final class HttpServerExchange extends AbstractAttachable {
      * @return A mutable map of response cookies
      * @deprecated use either {@link #responseCookies()} or {@link #setResponseCookie(Cookie)} methods instead
      */
-    @Deprecated
+    @Deprecated(since="2.2.0", forRemoval=true)
     public Map<String, Cookie> getResponseCookies() {
         if (deprecatedResponseCookies == null) {
             deprecatedResponseCookies = new MapDelegatingToSet((Set<Cookie>)((DelegatingIterable<Cookie>)responseCookies()).getDelegate());
@@ -1334,6 +1341,7 @@ public final class HttpServerExchange extends AbstractAttachable {
             return;
         }
         if (requestChannel != null) {
+            requestChannel.suspendReads();
             requestChannel.requestDone();
         }
         this.state = oldVal | FLAG_REQUEST_TERMINATED;
@@ -1630,6 +1638,7 @@ public final class HttpServerExchange extends AbstractAttachable {
             return this;
         }
         if(responseChannel != null) {
+            responseChannel.suspendWrites();
             responseChannel.responseDone();
         }
         this.state = oldVal | FLAG_RESPONSE_TERMINATED;
@@ -2014,9 +2023,15 @@ public final class HttpServerExchange extends AbstractAttachable {
         @Override
         public void close() throws IOException {
             try {
-                getInputStream().close();
+                // close the input stream if open or force closing if the request is incomplete
+                if (inputStream != null || !exchange.isRequestComplete()) {
+                    getInputStream().close();
+                }
             } finally {
-                getOutputStream().close();
+                // same for the output stream
+                if (outputStream != null || !exchange.isResponseComplete()) {
+                    getOutputStream().close();
+                }
             }
         }
 
