@@ -42,6 +42,7 @@ import io.undertow.util.LocaleUtils;
  * Handler that dumps a exchange to a log.
  *
  * @author Stuart Douglas
+ * @author <a href="mailto:ropalka@redhat.com">Richard Opalka</a>
  */
 public class RequestDumpingHandler implements HttpHandler {
 
@@ -61,7 +62,6 @@ public class RequestDumpingHandler implements HttpHandler {
         sb.append(" characterEncoding=" + exchange.getRequestHeaders().get(Headers.CONTENT_ENCODING) + "\n");
         sb.append("     contentLength=" + exchange.getRequestContentLength() + "\n");
         sb.append("       contentType=" + exchange.getRequestHeaders().get(Headers.CONTENT_TYPE) + "\n");
-        //sb.append("       contextPath=" + exchange.getContextPath());
         if (sc != null) {
             if (sc.isAuthenticated()) {
                 sb.append("          authType=" + sc.getMechanismName() + "\n");
@@ -71,13 +71,9 @@ public class RequestDumpingHandler implements HttpHandler {
             }
         }
 
-        Map<String, Cookie> cookies = exchange.getRequestCookies();
-        if (cookies != null) {
-            for (Map.Entry<String, Cookie> entry : cookies.entrySet()) {
-                Cookie cookie = entry.getValue();
-                sb.append("            cookie=" + cookie.getName() + "=" +
-                        cookie.getValue() + "\n");
-            }
+        for (Cookie cookie : exchange.requestCookies()) {
+            sb.append("            cookie=" + cookie.getName() + "=" +
+                    cookie.getValue() + "\n");
         }
         for (HeaderValues header : exchange.getRequestHeaders()) {
             for (String value : header) {
@@ -101,16 +97,13 @@ public class RequestDumpingHandler implements HttpHandler {
             }
             sb.append("\n");
         }
-        //sb.append("          pathInfo=" + exchange.getPathInfo());
         sb.append("          protocol=" + exchange.getProtocol() + "\n");
         sb.append("       queryString=" + exchange.getQueryString() + "\n");
         sb.append("        remoteAddr=" + exchange.getSourceAddress() + "\n");
         sb.append("        remoteHost=" + exchange.getSourceAddress().getHostName() + "\n");
-        //sb.append("requestedSessionId=" + exchange.getRequestedSessionId());
         sb.append("            scheme=" + exchange.getRequestScheme() + "\n");
         sb.append("              host=" + exchange.getRequestHeaders().getFirst(Headers.HOST) + "\n");
         sb.append("        serverPort=" + exchange.getDestinationAddress().getPort() + "\n");
-        //sb.append("       servletPath=" + exchange.getServletPath());
         sb.append("          isSecure=" + exchange.isSecure() + "\n");
 
         exchange.addExchangeCompleteListener(new ExchangeCompletionListener() {
@@ -131,11 +124,8 @@ public class RequestDumpingHandler implements HttpHandler {
                 }
                 sb.append("     contentLength=" + exchange.getResponseContentLength() + "\n");
                 sb.append("       contentType=" + exchange.getResponseHeaders().getFirst(Headers.CONTENT_TYPE) + "\n");
-                Map<String, Cookie> cookies = exchange.getResponseCookies();
-                if (cookies != null) {
-                    for (Cookie cookie : cookies.values()) {
-                        sb.append("            cookie=" + cookie.getName() + "=" + cookie.getValue() + "; domain=" + cookie.getDomain() + "; path=" + cookie.getPath() + "\n");
-                    }
+                for (Cookie cookie : exchange.responseCookies()) {
+                    sb.append("            cookie=" + cookie.getName() + "=" + cookie.getValue() + "; domain=" + cookie.getDomain() + "; path=" + cookie.getPath() + "\n");
                 }
                 for (HeaderValues header : exchange.getResponseHeaders()) {
                     for (String value : header) {
@@ -193,6 +183,11 @@ public class RequestDumpingHandler implements HttpHandler {
         }
     }
 
+    @Override
+    public String toString() {
+        return "dump-request()";
+    }
+
 
     public static class Builder implements HandlerBuilder {
 
