@@ -115,6 +115,8 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
     private volatile boolean receivesSuspendedTooManyQueuedMessages = false;
     private volatile boolean receivesSuspendedTooManyBuffers = false;
 
+    // freed method forces to use a different lock as it can be called from everywhere
+    private final Object lockTooManyQueuedMessages = new Object();
 
     @SuppressWarnings("unused")
     private volatile int readsBroken = 0;
@@ -160,22 +162,18 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
         @Override
         public void freed() {
             int res = outstandingBuffersUpdater.decrementAndGet(AbstractFramedChannel.this);
-            if (!receivesSuspendedByUser && res == maxQueuedBuffers - 1) {
-                //we need to do the resume in the IO thread, as there is a risk of deadlock otherwise, as the calling thread is an application thread
-                //and may hold a lock on a stream source channel, see UNDERTOW-1312
-                getIoThread().execute(new Runnable() {
-                    @Override
-                    public void run() {
-                        synchronized (AbstractFramedChannel.this) {
-                            if (outstandingBuffersUpdater.get(AbstractFramedChannel.this) < maxQueuedBuffers) {
-                                if (UndertowLogger.REQUEST_IO_LOGGER.isTraceEnabled()) {
-                                    UndertowLogger.REQUEST_IO_LOGGER.tracef("Resuming reads on %s as buffers have been consumed", AbstractFramedChannel.this);
-                                }
-                                new UpdateResumeState(null, false, null).run();
-                            }
+            // do not resume immediately, take some time to consume half the buffers
+            if (res == maxQueuedBuffers / 2) {
+                synchronized (lockTooManyQueuedMessages) {
+                    if (receivesSuspendedTooManyBuffers) {
+                        if (UndertowLogger.REQUEST_IO_LOGGER.isTraceEnabled()) {
+                            UndertowLogger.REQUEST_IO_LOGGER.tracef("Resuming reads on %s as buffers have been consumed", AbstractFramedChannel.this);
                         }
+                        //we need to do the resume in the IO thread, as there is a risk of deadlock otherwise, as the calling thread is an application thread
+                        //and may hold a lock on a stream source channel, see UNDERTOW-1312, resume and var update should be delayed to the IO thread
+                        runInIoThread(new UpdateResumeState(null, false, null));
                     }
-                });
+                }
             }
         }
     };
@@ -207,7 +205,7 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
      */
     protected AbstractFramedChannel(final StreamConnection connectedStreamChannel, ByteBufferPool bufferPool, FramePriority<C, R, S> framePriority, final PooledByteBuffer readData, OptionMap settings) {
         this.framePriority = framePriority;
-        this.maxQueuedBuffers = settings.get(UndertowOptions.MAX_QUEUED_READ_BUFFERS, 10);
+        this.maxQueuedBuffers = settings.get(UndertowOptions.MAX_QUEUED_READ_BUFFERS, 16);
         this.settings = settings;
         if (readData != null) {
             if(readData.getBuffer().hasRemaining()) {
@@ -247,6 +245,10 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
         return new IdleTimeoutConduit(connectedStreamChannel);
     }
 
+    /**
+     * Adds the task to the queue and executes the IO thread runner.
+     * @param task The task to be executed in the IO thread
+     */
     void runInIoThread(Runnable task) {
         this.taskRunQueue.add(task);
         try {
@@ -257,6 +259,19 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
         }
     }
 
+    /**
+     * Executes the task in IO thread now if the current thread is the IO thread
+     * or schedules it calling method <em>runInIoThread</em>.
+     * @param task The task to be executed in the IO thread
+     */
+    void runNowOrInIoThread(Runnable task) {
+        if (getIoThread() == Thread.currentThread()) {
+            task.run();
+        } else {
+            runInIoThread(task);
+        }
+    }
+
     /**
      * Get the buffer pool for this connection.
      *
@@ -335,192 +350,180 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
     }
 
     /**
-     * receive method, returns null if no frame is ready. Otherwise returns a
+     * Receive method, returns null if no frame is ready. Otherwise returns a
      * channel that can be used to read the frame contents.
      * <p>
      * Calling this method can also have the side effect of making additional data available to
      * existing source channels. In general if you suspend receives or don't have some other way
      * of calling this method then it can prevent frame channels for being fully consumed.
      */
-    public R receive() throws IOException {
-        // store in a local variable to prevent invoking lastDataRead twice
-        boolean receivedMinusOne = false;
-        try {
-            synchronized (this) {
-                if (readChannelDone && receiver == null) {
-                    //we have received the last frame, we just shut down and return
-                    //it would probably make more sense to have the last channel responsible for this
-                    //however it is much simpler just to have it here
-                    if(readData != null) {
-                        readData.close();
-                        readData = null;
-                    }
-                    channel.getSourceChannel().suspendReads();
-                    channel.getSourceChannel().shutdownReads();
-                    return null;
-                }
-                partialRead = false;
-                boolean requiresReinvoke = false;
-                int reinvokeDataRemaining = 0;
-                ReferenceCountedPooled pooled = this.readData;
-                boolean hasData = false;
+    public synchronized R receive() throws IOException {
+        ReferenceCountedPooled pooled = this.readData;
+        if (readChannelDone && receiver == null) {
+            //we have received the last frame, we just shut down and return
+            //it would probably make more sense to have the last channel responsible for this
+            //however it is much simpler just to have it here
+            if(pooled != null) {
+                pooled.close();
+                readData = null;
+            }
+            channel.getSourceChannel().suspendReads();
+            channel.getSourceChannel().shutdownReads();
+            return null;
+        }
+        partialRead = false;
+        boolean requiresReinvoke = false;
+        int reinvokeDataRemaining = 0;
+        boolean hasData = false;
+        if (pooled == null) {
+            pooled = allocateReferenceCountedBuffer();
+            if (pooled == null) {
+                return null;
+            }
+        } else if(pooled.isFreed()) {
+            //we attempt to re-used an existing buffer
+            if(!pooled.tryUnfree()) {
+                pooled = allocateReferenceCountedBuffer();
                 if (pooled == null) {
-                    pooled = allocateReferenceCountedBuffer();
-                    if (pooled == null) {
-                        return null;
-                    }
-                } else if(pooled.isFreed()) {
-                    //we attempt to re-used an existing buffer
-                    if(!pooled.tryUnfree()) {
-                        pooled = allocateReferenceCountedBuffer();
-                        if (pooled == null) {
-                            return null;
-                        }
-                    }
-                    pooled.getBuffer().clear();
-                } else {
-                    hasData = pooled.getBuffer().hasRemaining();
-                    pooled.getBuffer().compact();
-                }
-                boolean forceFree = false;
-                int read = 0;
-                try {
-                    read = channel.getSourceChannel().read(pooled.getBuffer());
-                    if (read == 0 && !hasData) {
-                        //no data, we just free the buffer
-                        forceFree = true;
-                        return null;
-                    } else if (read == -1 && !hasData) {
-                        forceFree = true;
-                        receivedMinusOne = readChannelDone = true;
-                        return null;
-                    } else if(isLastFrameReceived() && frameDataRemaining == 0) {
-                        //we got data, although we should have received the last frame
-                        forceFree = true;
-                        markReadsBroken(new ClosedChannelException());
-                    }
-                    pooled.getBuffer().flip();
-                    if(read == -1) {
-                        requiresReinvoke = true;
-                        reinvokeDataRemaining = pooled.getBuffer().remaining();
-                    }
-                    if (frameDataRemaining > 0) {
-                        if (frameDataRemaining >= pooled.getBuffer().remaining()) {
-                            frameDataRemaining -= pooled.getBuffer().remaining();
-                            if(receiver != null) {
-                                //we still create a pooled view, this means that if the buffer is still active we can re-used it
-                                //which prevents attacks based on sending lots of small fragments
-                                PooledByteBuffer frameData = pooled.createView();
-                                receiver.dataReady(null, frameData);
-                            } else {
-                                //we are dropping a frame
-                                pooled.close();
-                                readData = null;
-                            }
-                            if(frameDataRemaining == 0) {
-                                receiver = null;
-                            }
-                            return null;
-                        } else {
-                            PooledByteBuffer frameData = pooled.createView((int) frameDataRemaining);
-                            frameDataRemaining = 0;
-                            if(receiver != null) {
-                                receiver.dataReady(null, frameData);
-                            } else{
-                                //we are dropping the frame
-                                frameData.close();
-                            }
-                            receiver = null;
-                        }
-                        //if we read data into a frame we just return immediately, even if there is more remaining
-                        //see https://issues.jboss.org/browse/UNDERTOW-410
-                        //basically if we don't do this we loose some message ordering semantics
-                        //as the second message may be processed before the first one
-
-                        //this is problematic for HTTPS, where the read listener may also be invoked by a queued task
-                        //and not by the selector mechanism
-                        return null;
-                    }
-                    FrameHeaderData data = parseFrame(pooled.getBuffer());
-                    if (data != null) {
-                        PooledByteBuffer frameData;
-                        if (data.getFrameLength() >= pooled.getBuffer().remaining()) {
-                            frameDataRemaining = data.getFrameLength() - pooled.getBuffer().remaining();
-                            frameData = pooled.createView();
-                            pooled.getBuffer().position(pooled.getBuffer().limit());
-                        } else {
-                            frameData = pooled.createView((int) data.getFrameLength());
-                        }
-                        AbstractFramedStreamSourceChannel<?, ?, ?> existing = data.getExistingChannel();
-                        if (existing != null) {
-                            if (data.getFrameLength() > frameData.getBuffer().remaining()) {
-                                receiver = (R) existing;
-                            }
-                            existing.dataReady(data, frameData);
-                            if (isLastFrameReceived()) {
-                                handleLastFrame(existing);
-                            }
-                            return null;
-                        } else {
-                            boolean moreData = data.getFrameLength() > frameData.getBuffer().remaining();
-                            R newChannel = createChannel(data, frameData);
-                            if (newChannel != null) {
-                                if (moreData) {
-                                    receiver = newChannel;
-                                }
-
-                                if(isLastFrameReceived()) {
-                                    handleLastFrame(newChannel);
-                                }
-                            } else {
-                                frameData.close();
-                            }
-                            return newChannel;
-                        }
-                    } else {
-                        //we set partial read to true so the read listener knows not to immediately call receive again
-                        partialRead = true;
-                    }
                     return null;
-                } catch (IOException|RuntimeException|Error e) {
-                    //something has code wrong with parsing, close the read side
-                    //we don't close the write side, as the underlying implementation will most likely want to send an error
-                    markReadsBroken(e);
-                    forceFree = true;
-                    throw e;
-                }finally {
-                    //if the receive caused the channel to break the close listener may be have been called
-                    //which will make readData null
-                    if (readData != null) {
-                        if (!pooled.getBuffer().hasRemaining() || forceFree) {
-                            if(pooled.getBuffer().capacity() < 1024 || forceFree) {
-                                //if there is less than 1k left we don't allow it to be re-aquired
-                                readData = null;
-                            }
-                            //even though this is freed we may un-free it if we get a new packet
-                            //this prevents many small reads resulting in a large number of allocated buffers
-                            pooled.close();
-
-                        }
-                    }
-                    if(requiresReinvoke) {
-                        if(readData != null && !readData.isFreed()) {
-                            if(readData.getBuffer().remaining() == reinvokeDataRemaining) {
-                                readData.close();
-                                readData = null;
-                                UndertowLogger.REQUEST_IO_LOGGER.debugf("Partial message read before connection close %s", this);
-                            }
-                        }
-                        channel.getSourceChannel().wakeupReads();
-                    }
                 }
             }
-        } finally {
-            // read receivedMinusOne, and not readChannelDone
-            // to prevent lastDataRead being invoked twice in case of
-            // two concurrent receive invocations
-            if (receivedMinusOne) {
+            pooled.getBuffer().clear();
+        } else {
+            hasData = pooled.getBuffer().hasRemaining();
+            pooled.getBuffer().compact();
+        }
+        boolean forceFree = false;
+        int read = 0;
+        try {
+            read = channel.getSourceChannel().read(pooled.getBuffer());
+            if (read == 0 && !hasData) {
+                //no data, we just free the buffer
+                forceFree = true;
+                return null;
+            } else if (read == -1 && !hasData) {
+                forceFree = true;
+                readChannelDone = true;
                 lastDataRead();
+                return null;
+            } else if(isLastFrameReceived() && frameDataRemaining == 0) {
+                //we got data, although we should have received the last frame
+                forceFree = true;
+                markReadsBroken(new ClosedChannelException());
+            }
+            pooled.getBuffer().flip();
+            if(read == -1) {
+                requiresReinvoke = true;
+                reinvokeDataRemaining = pooled.getBuffer().remaining();
+            }
+            if (frameDataRemaining > 0) {
+                if (frameDataRemaining >= pooled.getBuffer().remaining()) {
+                    frameDataRemaining -= pooled.getBuffer().remaining();
+                    if(receiver != null) {
+                        //we still create a pooled view, this means that if the buffer is still active we can re-used it
+                        //which prevents attacks based on sending lots of small fragments
+                        PooledByteBuffer frameData = pooled.createView();
+                        receiver.dataReady(null, frameData);
+                    } else {
+                        //we are dropping a frame
+                        pooled.close();
+                        readData = null;
+                    }
+                    if(frameDataRemaining == 0) {
+                        receiver = null;
+                    }
+                    return null;
+                } else {
+                    PooledByteBuffer frameData = pooled.createView((int) frameDataRemaining);
+                    frameDataRemaining = 0;
+                    if(receiver != null) {
+                        receiver.dataReady(null, frameData);
+                    } else{
+                        //we are dropping the frame
+                        frameData.close();
+                    }
+                    receiver = null;
+                }
+                //if we read data into a frame we just return immediately, even if there is more remaining
+                //see https://issues.jboss.org/browse/UNDERTOW-410
+                //basically if we don't do this we loose some message ordering semantics
+                //as the second message may be processed before the first one
+
+                //this is problematic for HTTPS, where the read listener may also be invoked by a queued task
+                //and not by the selector mechanism
+                return null;
+            }
+            FrameHeaderData data = parseFrame(pooled.getBuffer());
+            if (data != null) {
+                PooledByteBuffer frameData;
+                if (data.getFrameLength() >= pooled.getBuffer().remaining()) {
+                    frameDataRemaining = data.getFrameLength() - pooled.getBuffer().remaining();
+                    frameData = pooled.createView();
+                    pooled.getBuffer().position(pooled.getBuffer().limit());
+                } else {
+                    frameData = pooled.createView((int) data.getFrameLength());
+                }
+                AbstractFramedStreamSourceChannel<?, ?, ?> existing = data.getExistingChannel();
+                if (existing != null) {
+                    if (data.getFrameLength() > frameData.getBuffer().remaining()) {
+                        receiver = (R) existing;
+                    }
+                    existing.dataReady(data, frameData);
+                    if(isLastFrameReceived()) {
+                        handleLastFrame(existing);
+                    }
+                    return null;
+                } else {
+                    boolean moreData = data.getFrameLength() > frameData.getBuffer().remaining();
+                    R newChannel = createChannel(data, frameData);
+                    if (newChannel != null) {
+                        if (moreData) {
+                            receiver = newChannel;
+                        }
+
+                        if(isLastFrameReceived()) {
+                            handleLastFrame(newChannel);
+                        }
+                    } else {
+                        frameData.close();
+                    }
+                    return newChannel;
+                }
+            } else {
+                //we set partial read to true so the read listener knows not to immediately call receive again
+                partialRead = true;
+            }
+            return null;
+        } catch (IOException|RuntimeException|Error e) {
+            //something has code wrong with parsing, close the read side
+            //we don't close the write side, as the underlying implementation will most likely want to send an error
+            markReadsBroken(e);
+            forceFree = true;
+            throw e;
+        }finally {
+            //if the receive caused the channel to break the close listener may be have been called
+            //which will make readData null
+            if (readData != null) {
+                if (!pooled.getBuffer().hasRemaining() || forceFree) {
+                    if(pooled.getBuffer().capacity() < 1024 || forceFree) {
+                        //if there is less than 1k left we don't allow it to be re-aquired
+                        readData = null;
+                    }
+                    //even though this is freed we may un-free it if we get a new packet
+                    //this prevents many small reads resulting in a large number of allocated buffers
+                    pooled.close();
+
+                }
+            }
+            if(requiresReinvoke) {
+                if(pooled != null && !pooled.isFreed()) {
+                    if(pooled.getBuffer().remaining() == reinvokeDataRemaining) {
+                        pooled.close();
+                        readData = null;
+                        UndertowLogger.REQUEST_IO_LOGGER.debugf("Partial message read before connection close %s", this);
+                    }
+                }
+                channel.getSourceChannel().wakeupReads();
             }
         }
     }
@@ -544,15 +547,21 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
             int expect;
             do {
                 expect = outstandingBuffersUpdater.get(this);
-                if (expect == maxQueuedBuffers) {
-                    synchronized (this) {
-                        //we need to re-read in a sync block, to prevent races
+                if (expect >= maxQueuedBuffers || receivesSuspendedTooManyBuffers) {
+                    synchronized (lockTooManyQueuedMessages) {
+                        // although this method is already synched on this by receive, we need to synch on lockTooManyQueuedMessages because of freed
+                        if (receivesSuspendedTooManyBuffers) {
+                            // suspend already sent to the IO thread
+                            return null;
+                        }
                         expect = outstandingBuffersUpdater.get(this);
-                        if (expect == maxQueuedBuffers) {
+                        if (expect >= maxQueuedBuffers) {
                             if (UndertowLogger.REQUEST_IO_LOGGER.isTraceEnabled()) {
                                 UndertowLogger.REQUEST_IO_LOGGER.tracef("Suspending reads on %s due to too many outstanding buffers", this);
                             }
-                            getIoThread().execute(new UpdateResumeState(null, true, null));
+                            // suspend is done asap setting the var to true now
+                            receivesSuspendedTooManyBuffers = true;
+                            runNowOrInIoThread(new UpdateResumeState(null, null, null));
                             return null;
                         }
                     }
@@ -617,6 +626,8 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
                         framePriority.frameAdded(frame, pendingFrames, heldFrames);
                     }
                 } else {
+                    // frame not ready for sending, add in held queue and start the timeout task
+                    frame.addReadyForFlushTask();
                     heldFrames.add(frame);
                 }
             }
@@ -709,12 +720,7 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
         } finally {
             flushingSenders = false;
             if(!newFrames.isEmpty()) {
-                runInIoThread(new Runnable() {
-                    @Override
-                    public void run() {
-                        flushSenders();
-                    }
-                });
+                runInIoThread(this::flushSenders);
             }
         }
     }
@@ -750,16 +756,7 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
 
     public void flush() {
         if (!flushingSenders) {
-            if(channel.getIoThread() == Thread.currentThread()) {
-                flushSenders();
-            } else {
-                runInIoThread(new Runnable() {
-                    @Override
-                    public void run() {
-                        flushSenders();
-                    }
-                });
-            }
+            runNowOrInIoThread(this::flushSenders);
         }
     }
 
@@ -797,20 +794,32 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
      * Suspend the receive of new frames via {@link #receive()}
      */
     public synchronized void suspendReceives() {
+        // suspend is done asap setting the suspended var to true now
+        if (receivesSuspendedByUser) {
+            // already suspended
+            return;
+        }
         receivesSuspendedByUser = true;
-        getIoThread().execute(new UpdateResumeState(true, null, null));
+        runNowOrInIoThread(new UpdateResumeState(null, null, null));
     }
 
     /**
      * Resume the receive of new frames via {@link #receive()}
      */
     public synchronized void resumeReceives() {
+        // resume is delayed to the IO thread but the var is set to false now
+        if (!receivesSuspendedByUser) {
+            // already resumed
+            return;
+        }
         receivesSuspendedByUser = false;
-        getIoThread().execute(new UpdateResumeState(false, null, null));
+        runInIoThread(new UpdateResumeState(null, null, null));
     }
 
     private void doResume() {
-        if (readData != null && !readData.isFreed()) {
+        //NOTE: this should not require syncing with below part
+        final ReferenceCountedPooled localReadData = this.readData;
+        if (localReadData != null && !localReadData.isFreed()) {
             channel.getSourceChannel().wakeupReads();
         } else {
             channel.getSourceChannel().resumeReads();
@@ -826,15 +835,14 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
      */
     @Override
     public void close() throws IOException {
-        synchronized (this) {
-            if (UndertowLogger.REQUEST_IO_LOGGER.isTraceEnabled()) {
-                UndertowLogger.REQUEST_IO_LOGGER.tracef(new ClosedChannelException(), "Channel %s is being closed", this);
-            }
-            safeClose(channel);
-            if (readData != null) {
-                readData.close();
-                readData = null;
-            }
+        if (UndertowLogger.REQUEST_IO_LOGGER.isTraceEnabled()) {
+            UndertowLogger.REQUEST_IO_LOGGER.tracef(new ClosedChannelException(), "Channel %s is being closed", this);
+        }
+        safeClose(channel);
+        final ReferenceCountedPooled localReadData = this.readData;
+        if (localReadData != null) {
+            localReadData.close();
+            this.readData = null;
         }
         closeSubChannels();
     }
@@ -962,24 +970,16 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
                 if (listener == null) {
                     listener = DRAIN_LISTENER;
                 }
-                UndertowLogger.REQUEST_IO_LOGGER.tracef("Invoking receive listener", receiver);
+                //UndertowLogger.REQUEST_IO_LOGGER.tracef("Invoking receive listener: %s - receiver: %s", listener, receiver);
                 ChannelListeners.invokeChannelListener(AbstractFramedChannel.this, listener);
             }
             final boolean partialRead;
             synchronized (AbstractFramedChannel.this) {
                 partialRead = AbstractFramedChannel.this.partialRead;
             }
-            if (readData != null && !readData.isFreed() && channel.isOpen() && !partialRead) {
-                try {
-                    runInIoThread(new Runnable() {
-                        @Override
-                        public void run() {
-                            ChannelListeners.invokeChannelListener(channel, FrameReadListener.this);
-                        }
-                    });
-                } catch (RejectedExecutionException e) {
-                    IoUtils.safeClose(AbstractFramedChannel.this);
-                }
+            final ReferenceCountedPooled localReadData = readData;
+            if (localReadData != null && !localReadData.isFreed() && channel.isOpen() && !partialRead) {
+                runInIoThread(() -> ChannelListeners.invokeChannelListener(channel, FrameReadListener.this));
             }
             synchronized (AbstractFramedChannel.this) {
                 AbstractFramedChannel.this.partialRead = false;
@@ -1009,30 +1009,26 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
         @Override
         public void handleEvent(final CloseableChannel c) {
             if (Thread.currentThread() != c.getIoThread() && !c.getWorker().isShutdown()) {
-                runInIoThread(new Runnable() {
-                    @Override
-                    public void run() {
-                        ChannelListeners.invokeChannelListener(c, FrameCloseListener.this);
-                    }
-                });
+                runInIoThread(() -> ChannelListeners.invokeChannelListener(c, FrameCloseListener.this));
                 return;
             }
 
-
             if(c instanceof  StreamSinkChannel) {
                 sinkClosed = true;
             } else if(c instanceof StreamSourceChannel) {
                 sourceClosed = true;
             }
+
+            final ReferenceCountedPooled localReadData = readData;
             if(!sourceClosed || !sinkClosed) {
                 return; //both sides need to be closed
-            } else if(readData != null && !readData.isFreed()) {
+            } else if(localReadData != null && !localReadData.isFreed()) {
                 //we make sure there is no data left to receive, if there is then we invoke the receive listener
                 runInIoThread(new Runnable() {
                     @Override
                     public void run() {
-                        while (readData != null  && !readData.isFreed()) {
-                            int rem = readData.getBuffer().remaining();
+                        while (localReadData != null  && !localReadData.isFreed()) {
+                            int rem = localReadData.getBuffer().remaining();
                             ChannelListener listener = receiveSetter.get();
                             if(listener == null) {
                                 listener = DRAIN_LISTENER;
@@ -1041,7 +1037,7 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
                             if(!AbstractFramedChannel.this.isOpen()) {
                                 break;
                             }
-                            if (readData != null && rem == readData.getBuffer().remaining()) {
+                            if (localReadData != null && rem == localReadData.getBuffer().remaining()) {
                                 break;//make sure we are making progress
                             }
                         }
@@ -1091,8 +1087,8 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
                 } finally {
                     synchronized (AbstractFramedChannel.this) {
                         closeSubChannels();
-                        if (readData != null) {
-                            readData.close();
+                        if (localReadData != null) {
+                            localReadData.close();
                             readData = null;
                         }
                     }
@@ -1122,7 +1118,24 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
 
     @Override
     public String toString() {
-        return getClass().getSimpleName() + " peer " + channel.getPeerAddress() + " local " + channel.getLocalAddress() + "[ " + (receiver == null ? "No Receiver" : receiver.toString()) + " " + pendingFrames.toString() + " -- " + heldFrames.toString() + " -- " + newFrames.toString() + "]";
+        final StringBuilder stringBuilder = new StringBuilder(150);
+        stringBuilder.append(getClass().getSimpleName())
+        .append(" peer ")
+        .append(channel.getPeerAddress())
+        .append(" local ")
+        .append(channel.getLocalAddress())
+        .append("[ ");
+        synchronized (this) {
+            stringBuilder.append((receiver == null ? "No Receiver" : receiver.toString()))
+            .append(" ")
+            .append(pendingFrames.toString())
+            .append(" -- ")
+            .append(heldFrames.toString())
+            .append(" -- ")
+            .append(newFrames.toString());
+        }
+
+        return stringBuilder.toString();
     }
 
     protected StreamConnection getUnderlyingConnection() {
@@ -1165,16 +1178,20 @@ public abstract class AbstractFramedChannel<C extends AbstractFramedChannel<C, R
 
         @Override
         public void run() {
-            if (user != null) {
-                receivesSuspendedByUser = user;
+            boolean suspend;
+            synchronized (AbstractFramedChannel.this) {
+                if (user != null) {
+                    receivesSuspendedByUser = user;
+                }
+                if (buffers != null) {
+                    receivesSuspendedTooManyBuffers = buffers;
+                }
+                if (frames != null) {
+                    receivesSuspendedTooManyQueuedMessages = frames;
+                }
+                suspend = receivesSuspendedByUser || receivesSuspendedTooManyQueuedMessages || receivesSuspendedTooManyBuffers;
             }
-            if (buffers != null) {
-                receivesSuspendedTooManyBuffers = buffers;
-            }
-            if (frames != null) {
-                receivesSuspendedTooManyQueuedMessages = frames;
-            }
-            if (receivesSuspendedByUser || receivesSuspendedTooManyQueuedMessages || receivesSuspendedTooManyBuffers) {
+            if (suspend) {
                 channel.getSourceChannel().suspendReads();
             } else {
                 doResume();
