@@ -36,6 +36,8 @@ import io.undertow.util.Methods;
 import io.undertow.util.Protocols;
 import io.undertow.util.URLUtils;
 import io.undertow.util.BadRequestException;
+import io.undertow.util.Cookies;
+
 import org.xnio.OptionMap;
 
 import static io.undertow.util.Headers.ACCEPT_CHARSET_STRING;
@@ -162,14 +164,16 @@ public abstract class HttpRequestParser {
 
     private final int maxParameters;
     private final int maxHeaders;
-    private final boolean allowEncodedSlash;
+    private final boolean slashDecodingFlag;
     private final boolean decode;
     private final String charset;
     private final int maxCachedHeaderSize;
     private final boolean allowUnescapedCharactersInUrl;
+    private final boolean allowIDLessMatrixParams;
 
     private static final boolean[] ALLOWED_TARGET_CHARACTER = new boolean[256];
 
+    private static final String ID_LESS_MATRIX_PARAMS_PROPERTY = "io.undertow.server.protocol.http.Parser.ID_LESS_MATRIX_PARAMS_PROPERTY";
     static {
         try {
             HTTP = "HTTP/1.".getBytes("ASCII");
@@ -201,14 +205,19 @@ public abstract class HttpRequestParser {
         }
     }
 
+    public static boolean isTargetCharacterAllowed(char c) {
+        return ALLOWED_TARGET_CHARACTER[c];
+    }
+
     public HttpRequestParser(OptionMap options) {
         maxParameters = options.get(UndertowOptions.MAX_PARAMETERS, UndertowOptions.DEFAULT_MAX_PARAMETERS);
         maxHeaders = options.get(UndertowOptions.MAX_HEADERS, UndertowOptions.DEFAULT_MAX_HEADERS);
-        allowEncodedSlash = options.get(UndertowOptions.ALLOW_ENCODED_SLASH, false);
+        slashDecodingFlag = URLUtils.getSlashDecodingFlag(options);
         decode = options.get(UndertowOptions.DECODE_URL, true);
         charset = options.get(UndertowOptions.URL_CHARSET, StandardCharsets.UTF_8.name());
         maxCachedHeaderSize = options.get(UndertowOptions.MAX_CACHED_HEADER_SIZE, UndertowOptions.DEFAULT_MAX_CACHED_HEADER_SIZE);
         this.allowUnescapedCharactersInUrl = options.get(UndertowOptions.ALLOW_UNESCAPED_CHARACTERS_IN_URL, false);
+        this.allowIDLessMatrixParams = Boolean.parseBoolean(System.getProperty(ID_LESS_MATRIX_PARAMS_PROPERTY));
     }
 
     public static final HttpRequestParser instance(final OptionMap options) {
@@ -289,9 +298,13 @@ public abstract class HttpRequestParser {
                     handleHeaderValue(buffer, currentState, builder);
                 }
             }
+            sanitazeListTypeHeaders(builder);
+            return;
+        } else {
+            handleStateful(buffer, currentState, builder);
+            sanitazeListTypeHeaders(builder);
             return;
         }
-        handleStateful(buffer, currentState, builder);
     }
 
     private void handleStateful(ByteBuffer buffer, ParseState currentState, HttpServerExchange builder) throws BadRequestException {
@@ -314,6 +327,13 @@ public abstract class HttpRequestParser {
             if (!buffer.hasRemaining()) {
                 return;
             }
+            // we could go back to PATH after PATH_PARAMETERS
+            if (currentState.state == ParseState.PATH) {
+                handlePath(buffer, currentState, builder);
+                if (!buffer.hasRemaining()) {
+                    return;
+                }
+            }
         }
 
         if (currentState.state == ParseState.VERSION) {
@@ -344,6 +364,13 @@ public abstract class HttpRequestParser {
         }
     }
 
+    private void sanitazeListTypeHeaders(final HttpServerExchange builder) {
+        //NOTE: this should be used only for HTTP1.x
+        if(!Cookies.isCrumbsAssemplyDisabled()) {
+            Cookies.assembleCrumbs(builder.getRequestHeaders());
+        }
+        //TODO: list type headers?
+    }
 
     abstract void handleHttpVerb(ByteBuffer buffer, final ParseState currentState, final HttpServerExchange builder) throws BadRequestException;
 
@@ -384,24 +411,9 @@ public abstract class HttpRequestParser {
             if (next == ' ' || next == '\t') {
                 if (stringBuilder.length() != 0) {
                     final String path = stringBuilder.toString();
-                    if(parseState == SECOND_SLASH) {
-                        exchange.setRequestPath("/");
-                        exchange.setRelativePath("/");
-                        exchange.setRequestURI(path);
-                    } else if (parseState < HOST_DONE) {
-                        String decodedPath = decode(path, urlDecodeRequired, state, allowEncodedSlash, false);
-                        exchange.setRequestPath(decodedPath);
-                        exchange.setRelativePath(decodedPath);
-                        exchange.setRequestURI(path);
-                    } else {
-                        handleFullUrl(state, exchange, canonicalPathStart, urlDecodeRequired, path);
-                    }
+                    parsePathComplete(state, exchange, canonicalPathStart, parseState, urlDecodeRequired, path);
                     exchange.setQueryString("");
                     state.state = ParseState.VERSION;
-                    state.stringBuilder.setLength(0);
-                    state.parseState = 0;
-                    state.pos = 0;
-                    state.urlDecodeRequired = false;
                     return;
                 }
             } else if (next == '\r' || next == '\n') {
@@ -409,10 +421,26 @@ public abstract class HttpRequestParser {
             } else if (next == '?' && (parseState == START || parseState == HOST_DONE || parseState == IN_PATH)) {
                 beginQueryParameters(buffer, state, exchange, stringBuilder, parseState, canonicalPathStart, urlDecodeRequired);
                 return;
-            } else if (next == ';' && (parseState == START || parseState == HOST_DONE || parseState == IN_PATH)) {
-                beginPathParameters(state, exchange, stringBuilder, parseState, canonicalPathStart, urlDecodeRequired);
+            } else if (next == ';') {
+                state.parseState = parseState;
+                state.urlDecodeRequired = urlDecodeRequired;
+                // store at canonical path the partial path parsed up until here
+                state.canonicalPath.append(stringBuilder.substring(canonicalPathStart));
+                state.stringBuilder.append(";");
+                // set position to end of path (possibly start of parameter name)
+                state.pos = state.stringBuilder.length();
+                // handle the path parameters
                 handlePathParameters(buffer, state, exchange);
-                return;
+                // if state is PATH, it means that handlePathParameters found a / after parsing path parameters
+                // so, we expect a next chunk of path in the next bytes, mark this with canonicalPathStart
+                // and append the '/' read by handlePathParameters
+                if(state.state == ParseState.PATH) {
+                    canonicalPathStart = stringBuilder.length();
+                    stringBuilder.append('/');
+                } else {
+                    // path has been entirely parsed, just return
+                    return;
+                }
             } else {
 
                 if (decode && (next == '%' || next > 127)) {
@@ -440,25 +468,21 @@ public abstract class HttpRequestParser {
         state.urlDecodeRequired = urlDecodeRequired;
     }
 
-    private void beginPathParameters(ParseState state, HttpServerExchange exchange, StringBuilder stringBuilder, int parseState, int canonicalPathStart, boolean urlDecodeRequired) {
-        final String path = stringBuilder.toString();
-        if(parseState == SECOND_SLASH) {
+    private void parsePathComplete(ParseState state, HttpServerExchange exchange, int canonicalPathStart, int parseState, boolean urlDecodeRequired, String path) {
+        if (parseState == SECOND_SLASH) {
             exchange.setRequestPath("/");
             exchange.setRelativePath("/");
-            exchange.setRequestURI(path);
-        } else if (parseState < HOST_DONE) {
-            String decodedPath = decode(path, urlDecodeRequired, state, allowEncodedSlash, false);
+            exchange.setRequestURI(path, true);
+        } else if (parseState < HOST_DONE && state.canonicalPath.length() == 0) {
+            String decodedPath = decode(path, urlDecodeRequired, state, slashDecodingFlag, false);
             exchange.setRequestPath(decodedPath);
             exchange.setRelativePath(decodedPath);
-            exchange.setRequestURI(path);
+            exchange.setRequestURI(path, false);
         } else {
-            String thePath = path.substring(canonicalPathStart);
-            exchange.setRequestPath(thePath);
-            exchange.setRelativePath(thePath);
-            exchange.setRequestURI(path, true);
+            handleFullUrl(state, exchange, canonicalPathStart, urlDecodeRequired, path, parseState);
         }
-        state.state = ParseState.PATH_PARAMETERS;
         state.stringBuilder.setLength(0);
+        state.canonicalPath.setLength(0);
         state.parseState = 0;
         state.pos = 0;
         state.urlDecodeRequired = false;
@@ -466,36 +490,22 @@ public abstract class HttpRequestParser {
 
     private void beginQueryParameters(ByteBuffer buffer, ParseState state, HttpServerExchange exchange, StringBuilder stringBuilder, int parseState, int canonicalPathStart, boolean urlDecodeRequired) throws BadRequestException {
         final String path = stringBuilder.toString();
-        if (parseState == SECOND_SLASH) {
-            exchange.setRequestPath("/");
-            exchange.setRelativePath("/");
-            exchange.setRequestURI(path);
-        } else if (parseState < HOST_DONE) {
-            String decodedPath = decode(path, urlDecodeRequired, state, allowEncodedSlash, false);
-            exchange.setRequestPath(decodedPath);
-            exchange.setRelativePath(decodedPath);
-            exchange.setRequestURI(path, false);
-        } else {
-            handleFullUrl(state, exchange, canonicalPathStart, urlDecodeRequired, path);
-        }
+        parsePathComplete(state, exchange, canonicalPathStart, parseState, urlDecodeRequired, path);
         state.state = ParseState.QUERY_PARAMETERS;
-        state.stringBuilder.setLength(0);
-        state.parseState = 0;
-        state.pos = 0;
-        state.urlDecodeRequired = false;
         handleQueryParameters(buffer, state, exchange);
     }
 
-    private void handleFullUrl(ParseState state, HttpServerExchange exchange, int canonicalPathStart, boolean urlDecodeRequired, String path) {
-        String thePath = decode(path.substring(canonicalPathStart), urlDecodeRequired, state, allowEncodedSlash, false);
+    private void handleFullUrl(ParseState state, HttpServerExchange exchange, int canonicalPathStart, boolean urlDecodeRequired, String path, int parseState) {
+        state.canonicalPath.append(path.substring(canonicalPathStart));
+        String thePath = decode(state.canonicalPath.toString(), urlDecodeRequired, state, slashDecodingFlag, false);
         exchange.setRequestPath(thePath);
         exchange.setRelativePath(thePath);
-        exchange.setRequestURI(path, true);
+        exchange.setRequestURI(path, parseState == HOST_DONE);
     }
 
 
     /**
-     * Parses a path value
+     * Parses query string parameters
      *
      * @param buffer   The buffer
      * @param state    The current state
@@ -577,21 +587,24 @@ public abstract class HttpRequestParser {
         state.mapCount = mapCount;
     }
 
-    private String decode(final String value, boolean urlDecodeRequired, ParseState state, final boolean allowEncodedSlash, final boolean formEncoded) {
+    private String decode(final String value, boolean urlDecodeRequired, ParseState state, final boolean slashDecodingFlag, final boolean formEncoded) {
         if (urlDecodeRequired) {
-            return URLUtils.decode(value, charset, allowEncodedSlash, formEncoded, state.decodeBuffer);
+            return URLUtils.decode(value, charset, slashDecodingFlag, formEncoded, state.decodeBuffer);
         } else {
             return value;
         }
     }
 
-
+    /**
+     * This method should only ever be called if the path contains am non-decoded semi colon character ';'
+     */
     final void handlePathParameters(ByteBuffer buffer, ParseState state, HttpServerExchange exchange) throws BadRequestException {
-        StringBuilder stringBuilder = state.stringBuilder;
-        int queryParamPos = state.pos;
-        int mapCount = state.mapCount;
+
+        state.state = ParseState.PATH_PARAMETERS;
         boolean urlDecodeRequired = state.urlDecodeRequired;
-        String nextQueryParam = state.nextQueryParam;
+        String param = state.nextQueryParam;
+        final StringBuilder stringBuilder = state.stringBuilder;
+        int pos = state.pos;
 
         //so this is a bit funky, because it not only deals with parsing, but
         //also deals with URL decoding the query parameters as well, while also
@@ -605,65 +618,65 @@ public abstract class HttpRequestParser {
             if(!allowUnescapedCharactersInUrl && !ALLOWED_TARGET_CHARACTER[next]) {
                 throw new BadRequestException(UndertowMessages.MESSAGES.invalidCharacterInRequestTarget(next));
             }
+            // end of HTTP URI
             if (next == ' ' || next == '\t' || next == '?') {
-                if (nextQueryParam == null) {
-                    if (queryParamPos != stringBuilder.length()) {
-                        exchange.addPathParam(decode(stringBuilder.substring(queryParamPos), urlDecodeRequired, state, true, true), "");
-                    }
-                } else {
-                    exchange.addPathParam(nextQueryParam, decode(stringBuilder.substring(queryParamPos), urlDecodeRequired, state, true, true));
-                }
-                exchange.setRequestURI(exchange.getRequestURI() + ';' + stringBuilder.toString(), state.parseState > HOST_DONE);
-                state.stringBuilder.setLength(0);
-                state.pos = 0;
+                handleParsedParam(param, stringBuilder.substring(pos), exchange, urlDecodeRequired, state);
+                final String path = stringBuilder.toString();
+                // the canonicalPathStart should be the current length to not add anything to it
+                parsePathComplete(state, exchange, path.length(), state.parseState, state.urlDecodeRequired, path);
+                state.state = ParseState.VERSION;
                 state.nextQueryParam = null;
-                state.mapCount = 0;
-                state.urlDecodeRequired = false;
                 if (next == '?') {
                     state.state = ParseState.QUERY_PARAMETERS;
                     handleQueryParameters(buffer, state, exchange);
-                } else {
-                    state.state = ParseState.VERSION;
-                }
+                } else
+                    exchange.setQueryString("");
                 return;
             } else if (next == '\r' || next == '\n') {
                 throw UndertowMessages.MESSAGES.failedToParsePath();
+            } else if (next == '/') {
+                handleParsedParam(param, stringBuilder.substring(pos), exchange, urlDecodeRequired, state);
+                state.pos = stringBuilder.length();
+                state.state = ParseState.PATH;
+                state.nextQueryParam = null;
+                return;
             } else {
                 if (decode && (next == '+' || next == '%' || next > 127)) {
                     urlDecodeRequired = true;
                 }
-                if (next == '=' && nextQueryParam == null) {
-                    nextQueryParam = decode(stringBuilder.substring(queryParamPos), urlDecodeRequired, state, true, true);
+                if ((next == '=' || (next == ',' && this.allowIDLessMatrixParams)) && param == null) {
+                    param = decode(stringBuilder.substring(pos), urlDecodeRequired, state, true, true);
                     urlDecodeRequired = false;
-                    queryParamPos = stringBuilder.length() + 1;
-                } else if (next == '&' && nextQueryParam == null) {
-                    if (++mapCount >= maxParameters) {
-                        throw UndertowMessages.MESSAGES.tooManyQueryParameters(maxParameters);
+                    pos = stringBuilder.length() + 1;
+                } else if (next == ';') {
+                    handleParsedParam(param, stringBuilder.substring(pos), exchange, urlDecodeRequired, state);
+                    param = null;
+                    pos = stringBuilder.length() + 1;
+                } else if (next == ',') {
+                    if(param == null) {
+                        throw UndertowMessages.MESSAGES.failedToParsePath();
+                    } else {
+                        handleParsedParam(param, stringBuilder.substring(pos), exchange, urlDecodeRequired, state);
+                        pos = stringBuilder.length() + 1;
                     }
-                    exchange.addPathParam(decode(stringBuilder.substring(queryParamPos), urlDecodeRequired, state, true, true), "");
-                    urlDecodeRequired = false;
-                    queryParamPos = stringBuilder.length() + 1;
-                } else if (next == '&') {
-                    if (++mapCount >= maxParameters) {
-                        throw UndertowMessages.MESSAGES.tooManyQueryParameters(maxParameters);
-                    }
-
-                    exchange.addPathParam(nextQueryParam, decode(stringBuilder.substring(queryParamPos), urlDecodeRequired, state, true, true));
-                    urlDecodeRequired = false;
-                    queryParamPos = stringBuilder.length() + 1;
-                    nextQueryParam = null;
                 }
                 stringBuilder.append(next);
-
             }
-
         }
-        state.pos = queryParamPos;
-        state.nextQueryParam = nextQueryParam;
-        state.mapCount = mapCount;
+
         state.urlDecodeRequired = urlDecodeRequired;
+        state.pos = pos;
+        state.urlDecodeRequired = urlDecodeRequired;
+        state.nextQueryParam = param;
     }
 
+    private void handleParsedParam(String paramName, String parameterValue, HttpServerExchange exchange, boolean urlDecodeRequired, ParseState state) throws BadRequestException {
+        if (paramName == null) {
+            exchange.addPathParam(decode(parameterValue, urlDecodeRequired, state, true, true), "");
+        } else {  // path param already parsed so parse and add the value
+            exchange.addPathParam(paramName, decode(parameterValue, urlDecodeRequired, state, true, true));
+        }
+    }
 
     /**
      * The parse states for parsing heading values
