@@ -74,9 +74,8 @@ public class Http2UpgradeHandler implements HttpHandler {
     public void handleRequest(HttpServerExchange exchange) throws Exception {
         final String upgrade = exchange.getRequestHeaders().getFirst(Headers.UPGRADE);
         final String settings = exchange.getRequestHeaders().getFirst("HTTP2-Settings");
-        if(settings != null && upgrade != null
-                && upgradeStrings.contains(upgrade)) {
-            if(HttpContinue.requiresContinueResponse(exchange) && false) {
+        if(settings != null && upgrade != null && upgradeStrings.contains(upgrade)) {
+            if(HttpContinue.requiresContinueResponse(exchange)) {
                 HttpContinue.sendContinueResponse(exchange, new IoCallback() {
                     @Override
                     public void onComplete(HttpServerExchange exchange, Sender sender) {
