@@ -18,29 +18,27 @@
 
 package io.undertow.server;
 
-import java.io.IOException;
-import java.io.InputStream;
-import java.io.OutputStream;
-
+import io.undertow.UndertowOptions;
+import io.undertow.server.handlers.BlockingHandler;
+import io.undertow.testutils.DefaultServer;
+import io.undertow.testutils.HttpClientUtils;
+import io.undertow.testutils.HttpOneOnly;
+import io.undertow.testutils.ProxyIgnore;
+import io.undertow.testutils.TestHttpClient;
+import io.undertow.util.Headers;
 import io.undertow.util.StatusCodes;
 import org.apache.http.HttpResponse;
 import org.apache.http.client.methods.HttpPost;
 import org.apache.http.entity.StringEntity;
 import org.junit.Assert;
-import org.junit.Assume;
 import org.junit.BeforeClass;
 import org.junit.Test;
 import org.junit.runner.RunWith;
 import org.xnio.OptionMap;
 
-import io.undertow.UndertowOptions;
-import io.undertow.server.handlers.BlockingHandler;
-import io.undertow.testutils.DefaultServer;
-import io.undertow.testutils.HttpOneOnly;
-import io.undertow.testutils.HttpClientUtils;
-import io.undertow.testutils.ProxyIgnore;
-import io.undertow.testutils.TestHttpClient;
-import io.undertow.util.Headers;
+import java.io.IOException;
+import java.io.InputStream;
+import java.io.OutputStream;
 
 /**
  * @author Stuart Douglas
@@ -71,8 +69,6 @@ public class MaxRequestSizeTestCase {
 
     @Test
     public void testMaxRequestHeaderSize() throws IOException {
-        // FIXME UNDERTOW-1938 (returns 500 response instead of 200, sporadic)
-        Assume.assumeFalse(System.getProperty("os.name").startsWith("Windows"));
         OptionMap existing = DefaultServer.getUndertowOptions();
         final TestHttpClient client = new TestHttpClient();
         try {
