@@ -71,6 +71,31 @@ public class SimpleParserTestCase {
         Assert.assertEquals("/somepath%2fotherPath", result.getRequestURI());
     }
 
+    @Test
+    public void testEncodedSlashDisallowed_DECODE_FLAG() throws BadRequestException {
+        byte[] in = "GET /somepath%2FotherPath HTTP/1.1\r\n\r\n".getBytes();
+
+        final ParseState context = new ParseState(10);
+        HttpServerExchange result = new HttpServerExchange(null);
+        HttpRequestParser.instance(OptionMap.create(UndertowOptions.DECODE_SLASH, false)).handle(ByteBuffer.wrap(in), context, result);
+        Assert.assertSame(Methods.GET, result.getRequestMethod());
+        Assert.assertEquals("/somepath%2FotherPath", result.getRequestURI());
+        Assert.assertEquals("/somepath%2FotherPath", result.getRequestPath());
+    }
+
+    @Test
+    public void testEncodedSlashAllowed_DECODE_FLAG() throws BadRequestException {
+        byte[] in = "GET /somepath%2fotherPath HTTP/1.1\r\n\r\n".getBytes();
+
+        final ParseState context = new ParseState(10);
+        HttpServerExchange result = new HttpServerExchange(null);
+        //this also tests override of UndertowOptions.ALLOW_ENCODED_SLASH
+        HttpRequestParser.instance(OptionMap.create(UndertowOptions.ALLOW_ENCODED_SLASH, false, UndertowOptions.DECODE_SLASH, true)).handle(ByteBuffer.wrap(in), context, result);
+        Assert.assertSame(Methods.GET, result.getRequestMethod());
+        Assert.assertEquals("/somepath/otherPath", result.getRequestPath());
+        Assert.assertEquals("/somepath%2fotherPath", result.getRequestURI());
+    }
+
     @Test
     public void testColonSlashInURL() throws BadRequestException {
         byte[] in = "GET /a/http://myurl.com/b/c HTTP/1.1\r\n\r\n".getBytes();
@@ -419,6 +444,15 @@ public class SimpleParserTestCase {
         runTest(in);
     }
 
+    @Test
+    public void testDifferentCaseHeaders() throws BadRequestException {
+        final ParseState context = new ParseState(10);
+        HttpServerExchange result = new HttpServerExchange(null);
+        byte[] in = "GET /somepath HTTP/1.1\r\nHost: www.somehost.net\r\nhost: other\r\n\r\n".getBytes();
+        HttpRequestParser.instance(OptionMap.EMPTY).handle(ByteBuffer.wrap(in), context, result);
+        Assert.assertArrayEquals(result.getRequestHeaders().get("HOST").toArray(), new String[] {"www.somehost.net", "other"});
+    }
+
     @Test(expected = BadRequestException.class)
     public void testTabInsteadOfSpaceAfterVerb() throws BadRequestException {
         byte[] in = "GET\t/somepath HTTP/1.1\r\nHost:   www.somehost.net\r\nOtherHeader: some\r\n    value\r\n\r\n".getBytes();
@@ -540,6 +574,24 @@ public class SimpleParserTestCase {
 
     }
 
+    @Test
+    public void testQueryParams_DECODE_FLAG() throws BadRequestException {
+        byte[] in = "GET http://www.somehost.net/somepath?a=b%3e%2F&b=c&d&e&f= HTTP/1.1\nHost: \t www.somehost.net\nOtherHeader:\tsome\n \t value\n\r\n".getBytes();
+
+        final ParseState context = new ParseState(10);
+        HttpServerExchange result = new HttpServerExchange(null);
+        HttpRequestParser.instance(OptionMap.create(UndertowOptions.DECODE_SLASH, false)).handle(ByteBuffer.wrap(in), context, result);
+        Assert.assertEquals("/somepath", result.getRelativePath());
+        Assert.assertEquals("http://www.somehost.net/somepath", result.getRequestURI());
+        Assert.assertEquals("a=b%3e%2F&b=c&d&e&f=", result.getQueryString());
+        Assert.assertEquals("b>/", result.getQueryParameters().get("a").getFirst());
+        Assert.assertEquals("c", result.getQueryParameters().get("b").getFirst());
+        Assert.assertEquals("", result.getQueryParameters().get("d").getFirst());
+        Assert.assertEquals("", result.getQueryParameters().get("e").getFirst());
+        Assert.assertEquals("", result.getQueryParameters().get("f").getFirst());
+
+    }
+
     @Test
     public void testSameHttpStringReturned() throws BadRequestException {
         byte[] in = "GET http://www.somehost.net/somepath HTTP/1.1\nHost: \t www.somehost.net\nAccept-Charset:\tsome\n \t  value\n\r\n".getBytes();
