@@ -49,7 +49,6 @@ import io.undertow.UndertowOptions;
 import io.undertow.connector.ByteBufferPool;
 import io.undertow.protocols.alpn.ALPNManager;
 import io.undertow.protocols.alpn.ALPNProvider;
-import io.undertow.protocols.alpn.JettyAlpnProvider;
 import io.undertow.protocols.ssl.SNIContextMatcher;
 import io.undertow.protocols.ssl.SNISSLContext;
 import io.undertow.protocols.ssl.UndertowXnioSsl;
@@ -951,6 +950,26 @@ public class DefaultServer extends BlockJUnit4ClassRunner {
             if (loadBalancingProxyClient != null) {
                 loadBalancingProxyClient.closeCurrentConnections();
             }
+            waitWorkerRunnableCycle(worker);
+        }
+    }
+
+    public static OptionMap getProxyOptions() {
+        if (proxyOpenListener != null) {
+            return proxyOpenListener.getUndertowOptions();
+        } else {
+            return null;
+        }
+    }
+
+    public static void setProxyOptions(final OptionMap options) {
+        OptionMap.Builder builder = OptionMap.builder().addAll(options);
+        builder = builder.set(UndertowOptions.BUFFER_PIPELINED_DATA, true);
+
+        if (proxyOpenListener != null) {
+            proxyOpenListener.setUndertowOptions(builder.getMap());
+            proxyOpenListener.closeConnections();
+            waitWorkerRunnableCycle(worker);
         }
     }
 
@@ -1031,11 +1050,7 @@ public class DefaultServer extends BlockJUnit4ClassRunner {
             //so we can't tell that ALPN is enabled or now
             SSLEngine engine = getClientSSLContext().createSSLEngine();
             ALPNProvider provider = ALPNManager.INSTANCE.getProvider(engine);
-            if (provider instanceof JettyAlpnProvider) {
-                alpnEnabled = System.getProperty("alpn-boot-string") != null;
-            } else {
-                alpnEnabled = provider != null;
-            }
+            alpnEnabled = provider != null;
         }
         return alpnEnabled;
     }
