@@ -21,7 +21,7 @@ package io.undertow.servlet;
 import java.io.IOException;
 import java.util.Date;
 import java.util.Set;
-import javax.servlet.UnavailableException;
+import jakarta.servlet.UnavailableException;
 
 import org.jboss.logging.BasicLogger;
 import org.jboss.logging.Logger;
@@ -135,4 +135,9 @@ public interface UndertowServletLogger extends BasicLogger {
 
     @Message(id = 15023, value = "This Context has been already destroyed")
     IllegalStateException contextDestroyed();
+
+    @LogMessage(level = ERROR)
+    @Message(id = 15024, value = "Servlet %s init() method in web application %s threw exception")
+    void failedToLoad(String servletName, String appName, @Cause Throwable t);
+
 }
