@@ -72,6 +72,7 @@ import io.undertow.servlet.api.ThreadSetupHandler;
 import io.undertow.servlet.api.WebResourceCollection;
 import io.undertow.servlet.handlers.CrawlerSessionManagerHandler;
 import io.undertow.servlet.handlers.RedirectDirHandler;
+import io.undertow.servlet.handlers.SendErrorPageHandler;
 import io.undertow.servlet.handlers.ServletDispatchingHandler;
 import io.undertow.servlet.handlers.ServletHandler;
 import io.undertow.servlet.handlers.ServletInitialHandler;
@@ -89,11 +90,11 @@ import io.undertow.servlet.spec.ServletContextImpl;
 import io.undertow.servlet.spec.SessionCookieConfigImpl;
 import io.undertow.util.MimeMappings;
 
-import javax.servlet.ServletContainerInitializer;
-import javax.servlet.ServletContext;
-import javax.servlet.ServletContextEvent;
-import javax.servlet.ServletContextListener;
-import javax.servlet.ServletException;
+import jakarta.servlet.ServletContainerInitializer;
+import jakarta.servlet.ServletContext;
+import jakarta.servlet.ServletContextEvent;
+import jakarta.servlet.ServletContextListener;
+import jakarta.servlet.ServletException;
 import java.io.File;
 import java.nio.charset.Charset;
 import java.util.ArrayList;
@@ -108,10 +109,10 @@ import java.util.ServiceLoader;
 import java.util.Set;
 import java.util.TreeMap;
 
-import static javax.servlet.http.HttpServletRequest.BASIC_AUTH;
-import static javax.servlet.http.HttpServletRequest.CLIENT_CERT_AUTH;
-import static javax.servlet.http.HttpServletRequest.DIGEST_AUTH;
-import static javax.servlet.http.HttpServletRequest.FORM_AUTH;
+import static jakarta.servlet.http.HttpServletRequest.BASIC_AUTH;
+import static jakarta.servlet.http.HttpServletRequest.CLIENT_CERT_AUTH;
+import static jakarta.servlet.http.HttpServletRequest.DIGEST_AUTH;
+import static jakarta.servlet.http.HttpServletRequest.FORM_AUTH;
 
 /**
  * The deployment manager. This manager is responsible for controlling the lifecycle of a servlet deployment.
@@ -187,7 +188,6 @@ public class DeploymentManagerImpl implements DeploymentManager {
                 @Override
                 public Void call(HttpServerExchange exchange, Object ignore) throws Exception {
                     final ApplicationListeners listeners = createListeners();
-                    listeners.start();
 
                     deployment.setApplicationListeners(listeners);
 
@@ -207,6 +207,8 @@ public class DeploymentManagerImpl implements DeploymentManager {
                         }
                     }
 
+                    listeners.start();
+
                     deployment.getSessionManager().registerSessionListener(new SessionListenerBridge(deployment, listeners, servletContext));
                     for(SessionListener listener : deploymentInfo.getSessionListeners()) {
                         deployment.getSessionManager().registerSessionListener(listener);
@@ -225,6 +227,7 @@ public class DeploymentManagerImpl implements DeploymentManager {
                         wrappedHandlers = new PredicateHandler(DispatcherTypePredicate.REQUEST, securityHandler, wrappedHandlers);
                     }
                     HttpHandler outerHandlers = wrapHandlers(wrappedHandlers, deploymentInfo.getOuterHandlerChainWrappers());
+                    outerHandlers = new SendErrorPageHandler(outerHandlers);
                     wrappedHandlers = new PredicateHandler(DispatcherTypePredicate.REQUEST, outerHandlers, wrappedHandlers);
                     wrappedHandlers = handleDevelopmentModePersistentSessions(wrappedHandlers, deploymentInfo, deployment.getSessionManager(), servletContext);
 
@@ -582,7 +585,11 @@ public class DeploymentManagerImpl implements DeploymentManager {
                     }
                     for (Map.Entry<Integer, List<ManagedServlet>> load : loadOnStartup.entrySet()) {
                         for (ManagedServlet servlet : load.getValue()) {
-                            servlet.createServlet();
+                            try {
+                                servlet.createServlet();
+                            } catch (ServletException e) {
+                                UndertowServletLogger.ROOT_LOGGER.failedToLoad(servlet.getServletInfo().getName(), deployment.getDeploymentInfo().getDeploymentName(), e.getCause());
+                            }
                         }
                     }
 
