@@ -31,8 +31,8 @@ import java.util.HashSet;
 import java.util.List;
 import java.util.Map;
 import java.util.Set;
-import javax.servlet.DispatcherType;
-import javax.servlet.http.MappingMatch;
+import jakarta.servlet.DispatcherType;
+import jakarta.servlet.http.MappingMatch;
 
 import io.undertow.UndertowLogger;
 import io.undertow.server.HandlerWrapper;
@@ -330,7 +330,7 @@ public class ServletPathMatches {
             //initialize the extension map. This contains all the filers in the noExtension map, plus
             //any filters that match the extension key
             for (String ext : extensionMatches) {
-                extension.put(ext, new EnumMap<DispatcherType, List<ManagedFilter>>(DispatcherType.class));
+                extension.put(ext, new EnumMap<>(DispatcherType.class));
             }
 
             //loop over all the filters, and add them to the appropriate map in the correct order
