@@ -28,6 +28,7 @@ import io.undertow.server.handlers.form.MultiPartParserDefinition;
 import io.undertow.server.protocol.http.HttpAttachments;
 import io.undertow.server.session.Session;
 import io.undertow.server.session.SessionConfig;
+import io.undertow.servlet.UndertowServletLogger;
 import io.undertow.servlet.UndertowServletMessages;
 import io.undertow.servlet.api.AuthorizationManager;
 import io.undertow.servlet.api.Deployment;
@@ -47,6 +48,25 @@ import io.undertow.util.HeaderValues;
 import io.undertow.util.Headers;
 import io.undertow.util.HttpString;
 import io.undertow.util.LocaleUtils;
+import jakarta.servlet.AsyncContext;
+import jakarta.servlet.DispatcherType;
+import jakarta.servlet.MultipartConfigElement;
+import jakarta.servlet.RequestDispatcher;
+import jakarta.servlet.ServletConnection;
+import jakarta.servlet.ServletException;
+import jakarta.servlet.ServletInputStream;
+import jakarta.servlet.ServletRequest;
+import jakarta.servlet.ServletRequestWrapper;
+import jakarta.servlet.ServletResponse;
+import jakarta.servlet.ServletResponseWrapper;
+import jakarta.servlet.http.Cookie;
+import jakarta.servlet.http.HttpServletMapping;
+import jakarta.servlet.http.HttpServletRequest;
+import jakarta.servlet.http.HttpServletResponse;
+import jakarta.servlet.http.HttpSession;
+import jakarta.servlet.http.HttpUpgradeHandler;
+import jakarta.servlet.http.Part;
+import jakarta.servlet.http.PushBuilder;
 
 import java.io.BufferedReader;
 import java.io.IOException;
@@ -72,24 +92,7 @@ import java.util.List;
 import java.util.Locale;
 import java.util.Map;
 import java.util.Set;
-import javax.servlet.AsyncContext;
-import javax.servlet.DispatcherType;
-import javax.servlet.MultipartConfigElement;
-import javax.servlet.RequestDispatcher;
-import javax.servlet.ServletException;
-import javax.servlet.ServletInputStream;
-import javax.servlet.ServletRequest;
-import javax.servlet.ServletRequestWrapper;
-import javax.servlet.ServletResponse;
-import javax.servlet.ServletResponseWrapper;
-import javax.servlet.http.Cookie;
-import javax.servlet.http.HttpServletMapping;
-import javax.servlet.http.HttpServletRequest;
-import javax.servlet.http.HttpServletResponse;
-import javax.servlet.http.HttpSession;
-import javax.servlet.http.HttpUpgradeHandler;
-import javax.servlet.http.Part;
-import javax.servlet.http.PushBuilder;
+
 
 /**
  * The http servlet request implementation. This class is not thread safe
@@ -234,41 +237,45 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
         ServletPathMatch match = src.getOriginalServletPathMatch();
         final DispatcherType dispatcherType = getDispatcherType();
         //UNDERTOW-1899 - ERROR is essentially forward operation
-        if(dispatcherType == DispatcherType.FORWARD || dispatcherType == DispatcherType.ERROR) {
+        if(dispatcherType == DispatcherType.FORWARD || dispatcherType == DispatcherType.ERROR || dispatcherType == DispatcherType.ASYNC || dispatcherType == DispatcherType.REQUEST) {
             match = src.getServletPathMatch();
         }
         String matchValue;
-        switch (match.getMappingMatch()) {
-            case EXACT:
-                matchValue = match.getMatched();
-                if(matchValue.startsWith("/")) {
-                    matchValue = matchValue.substring(1);
-                }
-                break;
-            case DEFAULT:
-            case CONTEXT_ROOT:
-                matchValue = "";
-                break;
-            case PATH:
-                matchValue = match.getRemaining();
-                if (matchValue == null) {
+        if (match != null) {
+            switch (match.getMappingMatch()) {
+                case EXACT:
+                    matchValue = match.getMatched();
+                    if (matchValue.startsWith("/")) {
+                        matchValue = matchValue.substring(1);
+                    }
+                    break;
+                case DEFAULT:
+                case CONTEXT_ROOT:
                     matchValue = "";
-                } else if (matchValue.startsWith("/")) {
-                    matchValue = matchValue.substring(1);
-                }
-                break;
-            case EXTENSION:
-                String matched = match.getMatched();
-                String matchString = match.getMatchString();
-                int startIndex = matched.startsWith("/") ? 1 : 0;
-                int endIndex = matched.length() - matchString.length() + 1;
-                matchValue = matched.substring(startIndex, endIndex);
-                break;
-            default:
-                matchValue = match.getRemaining();
+                    break;
+                case PATH:
+                    matchValue = match.getRemaining();
+                    if (matchValue == null) {
+                        matchValue = "";
+                    } else if (matchValue.startsWith("/")) {
+                        matchValue = matchValue.substring(1);
+                    }
+                    break;
+                case EXTENSION:
+                    String matched = match.getMatched();
+                    String matchString = match.getMatchString();
+                    int startIndex = matched.startsWith("/") ? 1 : 0;
+                    int endIndex = matched.length() - matchString.length() + 1;
+                    matchValue = matched.substring(startIndex, endIndex);
+                    break;
+                default:
+                    matchValue = match.getRemaining();
+            }
+            return new MappingImpl(matchValue, match.getMatchString(), match.getMappingMatch(), match.getServletChain().getManagedServlet().getServletInfo().getName());
+        }
+        else
+            return null;
         }
-        return new MappingImpl(matchValue, match.getMatchString(), match.getMappingMatch(), match.getServletChain().getManagedServlet().getServletInfo().getName());
-    }
 
     @Override
     public int getIntHeader(final String name) {
@@ -295,7 +302,7 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
 
     @Override
     public String getPathTranslated() {
-        return getRealPath(getPathInfo());
+        return servletContext.getRealPath(getPathInfo());
     }
 
     @Override
@@ -374,6 +381,12 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
         if (session == null) {
             throw UndertowServletMessages.MESSAGES.noSession();
         }
+        if (this.exchange.getAttachment(ServletRequestContext.ATTACHMENT_KEY).getServletResponse().isCommitted()) {
+            if (!this.servletContext.getDeployment().getDeploymentInfo().isOrphanSessionAllowed()) {
+                throw UndertowServletMessages.MESSAGES.sessionIdChangeAfterResponseCommittedNotAllowed();
+            }
+            UndertowServletLogger.REQUEST_LOGGER.debug("Servlet container configured to permit session identifier changes after response was committed. This can result in a memory leak if session has no timeout.");
+        }
         String oldId = session.getId();
         Session underlyingSession;
         if(System.getSecurityManager() == null) {
@@ -386,6 +399,22 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
         return newId;
     }
 
+    @Override
+    public String getRequestId() {
+        return exchange.getRequestId();
+    }
+
+    @Override
+    public String getProtocolRequestId() {
+        return exchange.getConnection().getProtocolRequestId();
+    }
+
+    @Override
+    public ServletConnection getServletConnection() {
+        String connectionId = Long.toString(exchange.getConnection().getId());
+        return new ServletConnectionImpl(connectionId, exchange.getProtocol().toString(), isSecure());
+    }
+
     @Override
     public String getRequestURI() {
         //we need the non-decoded string, which means we need to use exchange.getRequestURI()
@@ -453,11 +482,6 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
         return sessionCookieSource() == SessionConfig.SessionCookieSource.URL;
     }
 
-    @Override
-    public boolean isRequestedSessionIdFromUrl() {
-        return isRequestedSessionIdFromURL();
-    }
-
     @Override
     public boolean authenticate(final HttpServletResponse response) throws IOException, ServletException {
         if (response.isCommitted()) {
@@ -1003,11 +1027,6 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
         return servletContext.getRequestDispatcher(realPath);
     }
 
-    @Override
-    public String getRealPath(final String path) {
-        return servletContext.getRealPath(path);
-    }
-
     @Override
     public int getRemotePort() {
         return exchange.getSourceAddress().getPort();
