@@ -34,17 +34,18 @@ import java.util.Map;
 import java.util.Set;
 import java.util.function.Supplier;
 
-import javax.servlet.ServletException;
-import javax.servlet.ServletOutputStream;
-import javax.servlet.SessionTrackingMode;
-import javax.servlet.http.Cookie;
-import javax.servlet.http.HttpServletResponse;
-import javax.servlet.http.HttpSession;
+import jakarta.servlet.ServletException;
+import jakarta.servlet.ServletOutputStream;
+import jakarta.servlet.SessionTrackingMode;
+import jakarta.servlet.http.Cookie;
+import jakarta.servlet.http.HttpServletResponse;
+import jakarta.servlet.http.HttpSession;
 
 import io.undertow.UndertowLogger;
 import io.undertow.server.HttpServerExchange;
 import io.undertow.server.protocol.http.HttpAttachments;
 import io.undertow.servlet.UndertowServletMessages;
+import io.undertow.servlet.core.DefaultCharsetMapping;
 import io.undertow.servlet.handlers.ServletRequestContext;
 import io.undertow.util.CanonicalPathUtils;
 import io.undertow.util.DateUtils;
@@ -55,12 +56,16 @@ import io.undertow.util.HttpString;
 import io.undertow.util.Protocols;
 import io.undertow.util.RedirectBuilder;
 import io.undertow.util.StatusCodes;
+import org.xnio.Bits;
 
 import static io.undertow.util.URLUtils.isAbsoluteUrl;
 
 /**
+ * Implementation of {@code HttpServletResponse}.
+ *
  * @author Stuart Douglas
  * @author <a href="mailto:ropalka@redhat.com">Richard Opalka</a>
+ * @author Flavia Rainone
  */
 public final class HttpServletResponseImpl implements HttpServletResponse {
 
@@ -73,15 +78,22 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     private PrintWriter writer;
     private Integer bufferSize;
     private long contentLength = -1;
-    private boolean insideInclude = false;
+
+    private int flags = 0;
+    // response is inside include
+    private static final int INSIDE_INCLUDE_FLAG          = 1 << 0x00;
+    // response is done
+    private static final int RESPONSE_DONE_FLAG           = 1 << 0x01;
+    // ignored flush has been performed
+    private static final int IGNORED_FLUSH_PERFORMED_FLAG = 1 << 0x02;
+    // prevents anything to be added to response
+    private static final int TREAT_AS_COMMITTED_FLAG      = 1 << 0x03;
+    // indicates that we are closing the response and no further content should be written
+    private static final int CONTENT_FULLY_WRITTEN_FLAG   = 1 << 0x04;
+    //if a content type has been set either implicitly or implicitly
+    private static final int CHARSET_SET_FLAG             = 1 << 0x05;
+
     private Locale locale;
-    private boolean responseDone = false;
-
-    private boolean ignoredFlushPerformed = false;
-
-    private boolean treatAsCommitted = false;
-
-    private boolean charsetSet = false; //if a content type has been set either implicitly or implicitly
     private String contentType;
     private String charset;
     private Supplier<Map<String, String>> trailerSupplier;
@@ -98,7 +110,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
 
     @Override
     public void addCookie(final Cookie newCookie) {
-        if (insideInclude) {
+        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG)) {
             return;
         }
         final ServletCookieAdaptor servletCookieAdaptor = new ServletCookieAdaptor(newCookie);
@@ -113,19 +125,9 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         return exchange.getResponseHeaders().contains(name);
     }
 
-    @Override
-    public String encodeUrl(final String url) {
-        return encodeURL(url);
-    }
-
-    @Override
-    public String encodeRedirectUrl(final String url) {
-        return encodeRedirectURL(url);
-    }
-
     @Override
     public void sendError(final int sc, final String msg) throws IOException {
-        if(insideInclude) {
+        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG)) {
             //not 100% sure this is the correct action
             return;
         }
@@ -144,7 +146,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         exchange.setStatusCode(sc);
         if(src.isRunningInsideHandler()) {
             //all we do is set the error on the context, we handle it when the request is returned
-            treatAsCommitted = true;
+            flags |= TREAT_AS_COMMITTED_FLAG;
             src.setError(sc, msg);
         } else {
             //if the src is null there is no outer handler, as we are in an asnc request
@@ -157,7 +159,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         responseState = ResponseState.NONE;
         resetBuffer();
         exchange.getResponseHeaders().remove(Headers.CONTENT_LENGTH);
-        treatAsCommitted = false;
+        flags &= ~TREAT_AS_COMMITTED_FLAG;
         final String location = servletContext.getDeployment().getErrorPages().getErrorLocation(sc);
         if (location != null) {
             RequestDispatcherImpl requestDispatcher = new RequestDispatcherImpl(location, servletContext);
@@ -239,7 +241,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         if(name == null) {
             throw UndertowServletMessages.MESSAGES.headerNameWasNull();
         }
-        if (insideInclude || ignoredFlushPerformed) {
+        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG | IGNORED_FLUSH_PERFORMED_FLAG)) {
             return;
         }
         if(name.equals(Headers.CONTENT_TYPE)) {
@@ -261,7 +263,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         if(name == null) {
             throw UndertowServletMessages.MESSAGES.headerNameWasNull();
         }
-        if (insideInclude || ignoredFlushPerformed || treatAsCommitted) {
+        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG | IGNORED_FLUSH_PERFORMED_FLAG | TREAT_AS_COMMITTED_FLAG | CONTENT_FULLY_WRITTEN_FLAG)) {
             return;
         }
         if(name.equals(Headers.CONTENT_TYPE) && !exchange.getResponseHeaders().contains(Headers.CONTENT_TYPE)) {
@@ -283,7 +285,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
 
     @Override
     public void setStatus(final int sc) {
-        if (insideInclude || treatAsCommitted) {
+        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG | TREAT_AS_COMMITTED_FLAG)) {
             return;
         }
         if (responseStarted()) {
@@ -292,14 +294,6 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         exchange.setStatusCode(sc);
     }
 
-    @Override
-    public void setStatus(final int sc, final String sm) {
-        setStatus(sc);
-        if(!insideInclude && servletContext.getDeployment().getDeploymentInfo().isSendCustomReasonPhraseOnError()) {
-            exchange.setReasonPhrase(sm);
-        }
-    }
-
     @Override
     public int getStatus() {
         return exchange.getStatusCode();
@@ -349,7 +343,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     @Override
     public String getContentType() {
         if (contentType != null) {
-            if (charsetSet) {
+            if (Bits.anyAreSet(flags, CHARSET_SET_FLAG)) {
                 return contentType + ";charset=" + getCharacterEncoding();
             } else {
                 return contentType;
@@ -371,7 +365,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     @Override
     public PrintWriter getWriter() throws IOException {
         if (writer == null) {
-            if (!charsetSet) {
+            if (!Bits.anyAreSet(flags, CHARSET_SET_FLAG)) {
                 //servet 5.5
                 setCharacterEncoding(getCharacterEncoding());
             }
@@ -398,10 +392,14 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
 
     @Override
     public void setCharacterEncoding(final String charset) {
-        if (insideInclude || responseStarted() || writer != null || isCommitted()) {
+        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG) || responseStarted() || writer != null || isCommitted()) {
             return;
         }
-        charsetSet = charset != null;
+        if (charset != null) {
+            flags |= CHARSET_SET_FLAG;
+        } else {
+            flags &= ~CHARSET_SET_FLAG;
+        }
         this.charset = charset;
         if (contentType != null) {
             exchange.getResponseHeaders().put(Headers.CONTENT_TYPE, getContentType());
@@ -415,7 +413,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
 
     @Override
     public void setContentLengthLong(final long len) {
-        if (insideInclude || responseStarted()) {
+        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG) || responseStarted()) {
             return;
         }
         if(len >= 0) {
@@ -427,20 +425,24 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     }
 
     boolean isIgnoredFlushPerformed() {
-        return ignoredFlushPerformed;
+        return Bits.anyAreSet(flags, IGNORED_FLUSH_PERFORMED_FLAG);
     }
 
     void setIgnoredFlushPerformed(boolean ignoredFlushPerformed) {
-        this.ignoredFlushPerformed = ignoredFlushPerformed;
+        if (ignoredFlushPerformed) {
+            flags |= IGNORED_FLUSH_PERFORMED_FLAG;
+        } else {
+            flags &= ~IGNORED_FLUSH_PERFORMED_FLAG;
+        }
     }
 
     private boolean responseStarted() {
-        return exchange.isResponseStarted() || ignoredFlushPerformed || treatAsCommitted;
+        return exchange.isResponseStarted() || Bits.anyAreSet(flags, IGNORED_FLUSH_PERFORMED_FLAG | TREAT_AS_COMMITTED_FLAG);
     }
 
     @Override
     public void setContentType(final String type) {
-        if (type == null || insideInclude || responseStarted()) {
+        if (type == null || Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG) || responseStarted()) {
             return;
         }
         ContentTypeInfo ct = servletContext.parseContentType(type);
@@ -448,10 +450,10 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         boolean useCharset = false;
         if(ct.getCharset() != null && writer == null && !isCommitted()) {
             charset = ct.getCharset();
-            charsetSet = true;
+            flags |= CHARSET_SET_FLAG;
             useCharset = true;
         }
-        if(useCharset || !charsetSet) {
+        if(useCharset || !Bits.anyAreSet(flags, CHARSET_SET_FLAG)) {
             exchange.getResponseHeaders().put(Headers.CONTENT_TYPE, ct.getHeader());
         } else if(ct.getCharset() == null) {
             exchange.getResponseHeaders().put(Headers.CONTENT_TYPE, ct.getHeader() + "; charset=" + charset);
@@ -489,7 +491,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     }
 
     public void closeStreamAndWriter() throws IOException {
-        if(treatAsCommitted) {
+        if(Bits.anyAreSet(flags, TREAT_AS_COMMITTED_FLAG)) {
             return;
         }
         if (writer != null) {
@@ -542,28 +544,34 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         responseState = ResponseState.NONE;
         exchange.getResponseHeaders().clear();
         exchange.setStatusCode(StatusCodes.OK);
-        treatAsCommitted = false;
+        flags &= ~(TREAT_AS_COMMITTED_FLAG | CONTENT_FULLY_WRITTEN_FLAG);
     }
 
     @Override
     public void setLocale(final Locale loc) {
-        if (insideInclude || responseStarted()) {
+        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG) || responseStarted()) {
             return;
         }
         this.locale = loc;
         exchange.getResponseHeaders().put(Headers.CONTENT_LANGUAGE, loc.getLanguage() + "-" + loc.getCountry());
-        if (!charsetSet && writer == null) {
+        if (!Bits.anyAreSet(flags, CHARSET_SET_FLAG) && writer == null) {
             final Map<String, String> localeCharsetMapping = servletContext.getDeployment().getDeploymentInfo().getLocaleCharsetMapping();
-            // Match full language_country_variant first, then language_country,
-            // then language only
-            String charset = localeCharsetMapping.get(locale.toString());
-            if (charset == null) {
-                charset = localeCharsetMapping.get(locale.getLanguage() + "_"
-                        + locale.getCountry());
+            // first try DD provided mappings
+            String charset = null;
+            if (!localeCharsetMapping.isEmpty()) {
+                charset = localeCharsetMapping.get(locale.toString());
                 if (charset == null) {
-                    charset = localeCharsetMapping.get(locale.getLanguage());
+                    charset = localeCharsetMapping.get(locale.getLanguage() + "_"
+                            + locale.getCountry());
+                    if (charset == null) {
+                        charset = localeCharsetMapping.get(locale.getLanguage());
+                    }
                 }
             }
+            // if DD provided mapping failed then use default
+            if (charset == null) {
+                charset = DefaultCharsetMapping.INSTANCE.getCharset(loc);
+            }
             if (charset != null) {
                 this.charset = charset;
                 if (contentType != null) {
@@ -583,10 +591,10 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     }
 
     public void responseDone() {
-        if (responseDone || treatAsCommitted) {
+        if (Bits.anyAreSet(flags, RESPONSE_DONE_FLAG | TREAT_AS_COMMITTED_FLAG)) {
             return;
         }
-        responseDone = true;
+        flags |= RESPONSE_DONE_FLAG;
         try {
             closeStreamAndWriter();
         } catch (IOException e) {
@@ -597,11 +605,15 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     }
 
     public boolean isInsideInclude() {
-        return insideInclude;
+        return Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG);
     }
 
     public void setInsideInclude(final boolean insideInclude) {
-        this.insideInclude = insideInclude;
+        if (insideInclude) {
+            this.flags |= INSIDE_INCLUDE_FLAG;
+        } else {
+            this.flags &= ~INSIDE_INCLUDE_FLAG;
+        }
     }
 
     public void setServletContext(final ServletContextImpl servletContext) {
@@ -790,7 +802,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     }
 
     public boolean isTreatAsCommitted() {
-        return treatAsCommitted;
+        return Bits.anyAreSet(flags, TREAT_AS_COMMITTED_FLAG);
     }
 
     @Override
@@ -821,4 +833,11 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         return trailerSupplier;
     }
 
+    /**
+     * Marks this response as closed for writing extra bytes, including the addition of headers.
+     */
+    void setContentFullyWritten() {
+        this.flags |= CONTENT_FULLY_WRITTEN_FLAG;
+    }
+
 }
