@@ -26,10 +26,10 @@ import java.io.IOException;
 import java.nio.ByteBuffer;
 import java.nio.channels.FileChannel;
 import java.util.concurrent.atomic.AtomicIntegerFieldUpdater;
-import javax.servlet.DispatcherType;
-import javax.servlet.ServletOutputStream;
-import javax.servlet.ServletRequest;
-import javax.servlet.WriteListener;
+import jakarta.servlet.DispatcherType;
+import jakarta.servlet.ServletOutputStream;
+import jakarta.servlet.ServletRequest;
+import jakarta.servlet.WriteListener;
 
 import io.undertow.UndertowLogger;
 import org.xnio.Buffers;
@@ -44,6 +44,7 @@ import io.undertow.server.protocol.http.HttpAttachments;
 import io.undertow.servlet.UndertowServletMessages;
 import io.undertow.servlet.handlers.ServletRequestContext;
 import io.undertow.util.Headers;
+import io.undertow.util.Methods;
 
 /**
  * This stream essentially has two modes. When it is being used in standard blocking mode then
@@ -140,20 +141,21 @@ public class ServletOutputStreamImpl extends ServletOutputStream implements Buff
         if (len < 1) {
             return;
         }
+        int finalLength = (int) Math.min(len, remainingContentLength());
 
         if (listener == null) {
             ByteBuffer buffer = buffer();
-            if (buffer.remaining() < len) {
-                writeTooLargeForBuffer(b, off, len, buffer);
+            if (buffer.remaining() < finalLength) {
+                writeTooLargeForBuffer(b, off, finalLength, buffer);
             } else {
-                buffer.put(b, off, len);
+                buffer.put(b, off, finalLength);
                 if (buffer.remaining() == 0) {
                     writeBufferBlocking(false);
                 }
             }
-            updateWritten(len);
+            updateWritten(finalLength);
         } else {
-            writeAsync(b, off, len);
+            writeAsync(b, off, finalLength);
         }
     }
 
@@ -235,6 +237,7 @@ public class ServletOutputStreamImpl extends ServletOutputStream implements Buff
         if (anyAreClear(state, FLAG_READY)) {
             throw UndertowServletMessages.MESSAGES.streamNotReady();
         }
+        len = (int) Math.min(len, remainingContentLength());
         //even though we are in async mode we are still buffering
         try {
             ByteBuffer buffer = buffer();
@@ -284,6 +287,7 @@ public class ServletOutputStreamImpl extends ServletOutputStream implements Buff
         if (len < 1) {
             return;
         }
+        len = (int) Math.min(len, remainingContentLength());
 
         if (listener == null) {
             //if we have received the exact amount of content write it out in one go
@@ -369,29 +373,23 @@ public class ServletOutputStreamImpl extends ServletOutputStream implements Buff
         this.written += len;
         long contentLength = servletRequestContext.getOriginalResponse().getContentLength();
         if (contentLength != -1 && this.written >= contentLength) {
-            close();
+            servletRequestContext.getOriginalResponse().setContentFullyWritten();
         }
     }
 
-    void updateWrittenAsync(final long len) throws IOException {
+    long remainingContentLength() throws IOException {
+        final long contentLength = servletRequestContext.getOriginalResponse().getContentLength();
+        if (contentLength != -1) {
+            return contentLength - written;
+        }
+        return Long.MAX_VALUE;
+    }
+
+    void  updateWrittenAsync(final long len) throws IOException {
         this.written += len;
         long contentLength = servletRequestContext.getOriginalResponse().getContentLength();
         if (contentLength != -1 && this.written >= contentLength) {
-            setFlags(FLAG_CLOSED);
-            //if buffersToWrite is set we are already flushing
-            //so we don't have to do anything
-            if (buffersToWrite == null && pendingFile == null) {
-                if (flushBufferAsync(true)) {
-                    channel.shutdownWrites();
-                    setFlags(FLAG_DELEGATE_SHUTDOWN);
-                    channel.flush();
-                    if (pooledBuffer != null) {
-                        pooledBuffer.close();
-                        buffer = null;
-                        pooledBuffer = null;
-                    }
-                }
-            }
+            servletRequestContext.getOriginalResponse().setContentFullyWritten();
         }
     }
 
@@ -533,6 +531,7 @@ public class ServletOutputStreamImpl extends ServletOutputStream implements Buff
         if (anyAreSet(state, FLAG_CLOSED) || servletRequestContext.getOriginalResponse().isTreatAsCommitted()) {
             throw UndertowServletMessages.MESSAGES.streamIsClosed();
         }
+        final long remainingContentLength = remainingContentLength();
         if (listener == null) {
             if (buffer != null && buffer.position() != 0) {
                 writeBufferBlocking(false);
@@ -542,6 +541,9 @@ public class ServletOutputStreamImpl extends ServletOutputStream implements Buff
             }
             long position = source.position();
             long count = source.size() - position;
+            if (count > remainingContentLength) {
+                count = remainingContentLength;
+            }
             Channels.transferBlocking(channel, source, position, count);
             updateWritten(count);
         } else {
@@ -550,7 +552,7 @@ public class ServletOutputStreamImpl extends ServletOutputStream implements Buff
 
             long pos = 0;
             try {
-                long size = source.size();
+                long size = Math.min (source.size(), remainingContentLength);
                 pos = source.position();
 
                 while (size - pos > 0) {
@@ -578,12 +580,8 @@ public class ServletOutputStreamImpl extends ServletOutputStream implements Buff
         }
         buffer.flip();
         while (buffer.hasRemaining()) {
-            if (writeFinal) {
-                channel.writeFinal(buffer);
-            } else {
-                channel.write(buffer);
-            }
-            if (buffer.hasRemaining()) {
+            int result = writeFinal ? channel.writeFinal(buffer) : channel.write(buffer);
+            if (result == 0) {
                 channel.awaitWritable();
             }
         }
@@ -603,13 +601,14 @@ public class ServletOutputStreamImpl extends ServletOutputStream implements Buff
             if (anyAreSet(state, FLAG_CLOSED)) return;
             setFlags(FLAG_CLOSED);
             clearFlags(FLAG_READY);
-            if (allAreClear(state, FLAG_WRITE_STARTED) && channel == null && servletRequestContext.getOriginalResponse().getHeader(Headers.CONTENT_LENGTH_STRING) == null) {
+            if (allAreClear(state, FLAG_WRITE_STARTED) && channel == null) {
                 if (servletRequestContext.getOriginalResponse().getHeader(Headers.TRANSFER_ENCODING_STRING) == null
                         && servletRequestContext.getExchange().getAttachment(HttpAttachments.RESPONSE_TRAILER_SUPPLIER) == null
                         && servletRequestContext.getExchange().getAttachment(HttpAttachments.RESPONSE_TRAILERS) == null) {
-                    if (buffer == null) {
+                    final String contentLength = servletRequestContext.getOriginalResponse().getHeader(Headers.CONTENT_LENGTH_STRING);
+                    if (buffer == null && (contentLength == null || !Methods.HEAD_STRING.equals(servletRequestContext.getOriginalRequest().getMethod()))) {
                         servletRequestContext.getExchange().getResponseHeaders().put(Headers.CONTENT_LENGTH, "0");
-                    } else {
+                    } else if (buffer != null && contentLength == null) {
                         servletRequestContext.getExchange().getResponseHeaders().put(Headers.CONTENT_LENGTH, Integer.toString(buffer.position()));
                     }
                 }
@@ -675,11 +674,14 @@ public class ServletOutputStreamImpl extends ServletOutputStream implements Buff
             clearFlags(FLAG_READY);
             if (allAreClear(state, FLAG_WRITE_STARTED) && channel == null) {
 
-                if (servletRequestContext.getOriginalResponse().getHeader(Headers.TRANSFER_ENCODING_STRING) == null) {
-                    if (buffer == null) {
-                        servletRequestContext.getOriginalResponse().setHeader(Headers.CONTENT_LENGTH, "0");
-                    } else {
-                        servletRequestContext.getOriginalResponse().setHeader(Headers.CONTENT_LENGTH, Integer.toString(buffer.position()));
+                if (servletRequestContext.getOriginalResponse().getHeader(Headers.TRANSFER_ENCODING_STRING) == null
+                        && servletRequestContext.getExchange().getAttachment(HttpAttachments.RESPONSE_TRAILER_SUPPLIER) == null
+                        && servletRequestContext.getExchange().getAttachment(HttpAttachments.RESPONSE_TRAILERS) == null) {
+                    final String contentLength = servletRequestContext.getOriginalResponse().getHeader(Headers.CONTENT_LENGTH_STRING);
+                    if (buffer == null && (contentLength == null || !Methods.HEAD_STRING.equals(servletRequestContext.getOriginalRequest().getMethod()))) {
+                        servletRequestContext.getExchange().getResponseHeaders().put(Headers.CONTENT_LENGTH, "0");
+                    } else if (buffer != null && contentLength == null) {
+                        servletRequestContext.getExchange().getResponseHeaders().put(Headers.CONTENT_LENGTH, Integer.toString(buffer.position()));
                     }
                 }
             }
